import sys, re
from Bio import SeqIO
from contig_tools.file_parsing import import_contig_records 
from contig_tools.utility import get_logger

cov_pattern = re.compile("cov_([0-9.]+)")
logger = get_logger(__file__)

def filtered_contigs_generator(contigs, min_length = 500, min_coverage = None):
    total_contigs = 0
    contigs_kept = 0
    for contig in contigs:
        total_contigs = total_contigs + 1
        result = cov_pattern.search(contig.name)
        if not min_coverage and len(contig) >= min_length:
            contigs_kept = contigs_kept + 1
            yield contig
        elif min_coverage:
            if result:
                if float(result.group(1)) >= min_coverage and len(contig) >= min_length:
                    contigs_kept = contigs_kept + 1
                    yield contig
            else:
                logger.warning(f'No coverage encoded in contig name')
                if len(contig) >= min_length:
                    contigs_kept = contigs_kept + 1
                    yield contig

    print("Starting contigs: {0}\nContigs kept: {1}".format(total_contigs, contigs_kept))

def filter_contig_file(fasta_file_path, min_contig_length, min_contig_coverage, output_file = None):
    if not output_file:
        output_file = fasta_file_path.replace('fa', 'filter_gt_{0}bp_gt_{1}cov.fa'.format(min_contig_length, min_contig_coverage))
    contigs = import_contig_records(fasta_file_path)
    with open(output_file, 'w') as filtered_fasta:
            SeqIO.write(filtered_contigs_generator(contigs,
                min_length = min_contig_length,
                min_coverage = min_contig_coverage),
                filtered_fasta, 'fasta')
    print(f'Filtered contigs written to {output_file}')



from contig_tools.file_parsing import import_contig_records
from contig_tools.contig_metrics import get_contig_metrics
from contig_tools.utility import get_logger

import yaml

logger = get_logger(__file__)

def print_check_contig_metrics(fasta_file_path, conditions_yml_file):
    '''Print results of checking contig metrics'''
    contigs = import_contig_records(fasta_file_path)
    contig_metrics = get_contig_metrics(contigs)
    result, failure_reasons = check_contig_metrics(contig_metrics, conditions_yml_file)

    print(result)
    if not result:
        for reason in failure_reasons:
            print(reason)

def check_contig_metrics(contig_metrics, conditions_yml_file):
    '''Check metrics of contigs based on conditions defined in a yaml file '''


    # check each condition
    failure_reasons = []
    all_pass = True
    with open(conditions_yml_file) as conditions_yml:
        conditions = yaml.load(conditions_yml, Loader=yaml.FullLoader)
        for metric_name in conditions:
            try:
                metric_value = contig_metrics[metric_name]
                condition_value = conditions[metric_name]['condition_value']
                # possible conditions
                # greater than
                if conditions[metric_name]['condition_type'] == 'gt':
                    if not metric_value > condition_value:
                        all_pass = False
                        failure_reasons.append('{0}: {1} not > {2}'.format(metric_name, metric_value, condition_value))
                # less than
                elif conditions[metric_name]['condition_type'] == 'lt':
                    if not metric_value < condition_value:
                        all_pass = False
                        failure_reasons.append('{0}: {1} not < {2}'.format(metric_name, metric_value, condition_value))
                # less than and greater than
                elif conditions[metric_name]['condition_type'] == 'lt_gt':
                    if not (metric_value < condition_value[0] and metric_value > condition_value[1]):
                        all_pass = False
                        failure_reasons.append('{0}: not {1} > {2} > {3}'.format(metric_name,condition_value[0], metric_value, condition_value[1]))

            except KeyError as e:
                logger.error("No such key {0}. The available metrics are {1}".format(e, ", ".join(contig_metrics.keys())))

        if all_pass:
            return True, []
        else:
           return False, failure_reasons
    



from Bio import SeqIO
def import_contig_records(fasta_file_path):
    '''Make a list of Seq records from a fasta filepath''' 
    contigs = []
    for record in SeqIO.parse(fasta_file_path, 'fasta'):
        contigs.append(record)
    return contigs


import logging
import os
import colorlog
def get_logger(filepath, level=logging.DEBUG):
    # create logger
    logger = logging.getLogger(os.path.basename(filepath))
    logger.setLevel(level)
    # create console handler and set level to debug
    console = logging.StreamHandler()
    console.setLevel(level)
    # create formatter
    format_str = '%(asctime)s - %(levelname)-8s - %(message)s'
    date_format = '%Y-%m-%d %H:%M:%S'
    if os.isatty(2):
        cformat = '%(log_color)s' + format_str
        colors = {'DEBUG': 'reset',
                  'INFO': 'reset',
                  'WARNING': 'bold_yellow',
                  'ERROR': 'bold_red',
                  'CRITICAL': 'bold_red'}
        formatter = colorlog.ColoredFormatter(cformat, date_format,
                                              log_colors=colors)
    else:
        formatter = logging.Formatter(format_str, date_format)
    # add formatter to ch
    console.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(console)

    return logger


import setuptools

def readme():
    with open('README.md') as f:
        return f.read()

setuptools.setup(
    name='contig tools',
    version='0.3.7',
    description='Utility package to parse multi fasta files resulting from de novo assembly',
    long_description=readme(),
    long_description_content_type="text/markdown",
    author='Anthony Underwood',
    author_email='au3@sanger.ac.uk',
    license='MIT',
    packages=setuptools.find_packages(),
    scripts=['contig_tools/scripts/contig-tools'],
    install_requires=['biopython', 'colorlog', 'pyyaml', 'parmap', 'tqdm'],
    python_requires='>=3.7',
    test_suite='nose.collector',
    tests_require=['nose', 'coverage'],
    include_package_data=True,
    classifiers=[ 
        'Development Status :: 3 - Alpha', 
        'Intended Audience :: Science/Research', 
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ]
)


from contig_tools.file_parsing import import_contig_records
import json

def Nx(contig_lengths, x = 50):
    '''return Nx of a list of contig lengths'''
    contig_lengths.sort(reverse=1)
    Nx_threshold = sum(contig_lengths) * x / 100
    
    cumulative_length = 0
    for length in contig_lengths:
        cumulative_length += length
        if cumulative_length >= Nx_threshold:
            Nx = length
            return Nx

def get_contig_metrics(contigs):
    '''return a dictionary of contig metrics given a list of BioPython Seq objects'''
    contig_metrics = {}
    # make a list of contig lengths
    contig_lengths = [len(contig.seq) for contig in contigs]
    number_of_gcs = sum([sum(map(contig.seq.upper().count, ['G', 'C'])) for contig in contigs])
    contig_metrics['Number of contigs'] = len(contig_lengths)
    contig_metrics['Number of contigs > 500bp'] = len([filtered_length for filtered_length in contig_lengths if filtered_length > 500])
    contig_metrics['Total length'] = sum(contig_lengths)
    contig_metrics['%GC'] = number_of_gcs/sum(contig_lengths)*float(100)
    contig_metrics['Largest contig'] = max(contig_lengths)
    contig_metrics['N{0} score'.format(50)] = Nx(contig_lengths, x = 50)
    return contig_metrics

def print_contig_metrics(fasta_file_path, format='tsv'):
    '''Print out contig metrics given the path to a fasta file containing multiple contigs'''
    contigs = import_contig_records(fasta_file_path)
    contig_metrics = get_contig_metrics(contigs)
    if format == 'tsv':
        print('metric\tvalue')
        for metric in sorted(contig_metrics.keys()):
                print('{0}\t{1}'.format(metric, contig_metrics[metric]))
    elif format == 'json':
        print(json.dumps(contig_metrics, indent=2))




from Bio.Blast.Applications import NcbimakeblastdbCommandline
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast import NCBIXML
import tempfile
import os
import csv
import parmap


def calculate_chunk_size(num_tasks, num_parallel_processes, update_timeline_num = 4):
    """
    calculate the size of the number of tasks to pass in each 'chunk' to each parallele process
    """
    num_chunks, extra = divmod(num_tasks, num_parallel_processes * update_timeline_num)
    if extra:
        num_chunks += 1
    return int(num_chunks)


def parse_genome_file(genome_filepath, query_file, percent_id_threshold):
    """
    This function executes the series of steps necessary to query a genome with the query sequences and
    report matches, co-located or not

    Args:
        genome_filepath (str): Path to a single genome file
        query_file (str): Path to a multi fasta file containing two or more loci that could be co-located
        percent_id_threshold(float): threshold below which hits will not be reported

    Returns:
        tuple: first item is genome file prefix, second item are the parsed genome result metrics
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        blast_database_path = make_blast_database(genome_filepath, f'{tmpdirname}/blast_database')
        blast_result_path = blast_search(query_file, blast_database_path, tmpdirname)
        parsed_blast_result = parse_blast(blast_result_path, percent_id_threshold)
        genome_prefix = os.path.splitext(os.path.basename(genome_filepath))[0]
        return (genome_prefix, parsed_blast_result)


def find_co_located(query_file, output_file, percent_id_threshold, genome_file=None, genome_file_list=None, num_parallel_processes=1, write_only_co_located=False):
    """
    Function to find co located loci on one or more genomes

    Args:
        query_file (str): Path to a multi fasta file containing two or more loci that could be co-located
        output_file(str): path to file where the results will be written
        percent_id_threshold(float): threshold below which hits will not be reported
        genome_file (str): Path to a single genome file
        genome_file_list (str): Path to a text file with one or more genome filepaths. 1 / line
        num_parallel_processes (int): number of parallel processes to run
        write_only_co_located (boolean): whether to write to the output file just those genomes where the queries are colocated
    """
    if genome_file:
        genome_files = [genome_file]
    else:
        with open(genome_file_list) as f:
            genome_files = [line.strip() for line in f.readlines()]
    
    chunksize = calculate_chunk_size(len(genome_files), num_parallel_processes)

    parsed_blast_results = parmap.map(parse_genome_file,
                                      genome_files,
                                      query_file,
                                      percent_id_threshold,
                                      pm_processes=num_parallel_processes,
                                      pm_chunksize=chunksize,
                                      pm_pbar=True)
    parsed_blast_results_dict = { genome:result for (genome,result) in parsed_blast_results }

    write_result_table(parsed_blast_results_dict, output_file, write_only_co_located)



def make_blast_database(genome_file, outdir):
    """
    function to make blast database
    Args:
        genome_file (str): Path to a genome file from which the database will be made
        outdir (str): Path to where the blast database will be created including the database prefix
    
    Returns:
        str: path to the blast database, including prefix
    """
    cmd = NcbimakeblastdbCommandline(dbtype="nucl",
                                     input_file=genome_file,
                                     title='genome',
                                     out=outdir )
    cmd()
    return(outdir)
    
def blast_search(query_seq, blast_database, outdir, evalue=0.001):
    """
    Perform a blast search. The result will be written to an xml file in blast 5 format
    Args:
        query_seq (str): path to a query sequence
        blast_database (str): path to the blast database including prefix
        outdir (str): path to directory where the blast_result.xml file will be written
        evalue (float): the expect value (e) use for setting a threshold of which hits to report
    
    Return:
        str path to the blast result file
    """
    cmd = NcbiblastnCommandline(query=query_seq,
                                db=blast_database,
                                evalue=0.001,
                                out=f'{outdir}/blast_result.xml',
                                dust='no',
                                soft_masking='false',
                                outfmt=5)
    cmd()
    return(f'{outdir}/blast_result.xml')
    
def parse_blast(blast_output, percent_id_threshold = 95):
    """
    Parses a blast result with one or more records and reports metrics on hots that have a percent
    identity match equal to or greater than the percent_id_threshold
    Args:
        blast_output (str): path to a blast result file in blast 5 XML format
        percent_id_threshold(float): threshold below which hits will not be reported
    
    Returns:
        list: A list of blast metrics (one per record corresponding to each query) where the items are dicts with keys
        query, subject, subject_start, subject_end, percent_id
    """ 
    metrics = []
    with open(blast_output) as result_handle:
        blast_records = NCBIXML.parse(result_handle)
        for blast_record in blast_records:
            result_metrics = {'query': blast_record.query}
            if len(blast_record.alignments) > 0:
                alignment_metrics = get_alignment_metrics(blast_record.alignments[0])
                # test percent_id threshold
                alignment_length = alignment_metrics['subject_end'] - alignment_metrics['subject_start'] + 1
                percent_id = alignment_metrics['identities']/alignment_length*100
                if percent_id >= percent_id_threshold:
                    alignment_metrics['percent_id'] = round(percent_id,1)
                    del alignment_metrics['identities']
                    result_metrics.update(alignment_metrics)
            else:
                result_metrics.update({
                    'subject': 'NA',
                    'subject_start': 'NA',
                    'subject_end': 'NA',
                    'percent_id': 'NA',
                })

            metrics.append(result_metrics)
    return(metrics)


def get_alignment_metrics(alignment):
    """
    method to return alignment metrics
    Args:
    alignment (Bio.Blast.Record.Alignment): An alignment object

    Returns:
        dict: Dictionary of metrics with keys subject, identities, subject_start, subject_end
    """
    title = alignment.title.split(" ")[1]
    identities = sum([hsp.identities for hsp in alignment.hsps])
    subject_starts = [hsp.sbjct_start for hsp in alignment.hsps]
    subject_ends = [hsp.sbjct_end for hsp in alignment.hsps]
    
    subject_positions = subject_starts
    subject_positions.extend(subject_ends)
    subject_start = min(subject_positions)
    subject_end = max(subject_positions)
    return({"subject": title,
            "identities": identities,
            "subject_start": subject_start,
            "subject_end": subject_end})

    
def write_result_table(genome_results, output_file, write_only_co_located = False):
    """
    method to write genome_results data to table
    Args:
        genome_results (list): A list of genome results. Each item contains a lits of dicts with blast metrics for each query
        output_file (str): path to file where the results will be written
        write_only_co_located (bolean): whether to write to the output file just those genomes where the queries are colocated
    """
    with open(output_file, 'w') as f:
        csv_writer = csv.writer(f, delimiter='\t')
        # make field names based on the first item
        first_genome = list(genome_results.keys())[0]
        fieldnames = ['genome', 'co-located']
        fieldnames.extend([f'{key}_{index+1}' for index, result in enumerate(genome_results[first_genome]) for key in result.keys()])
        csv_writer.writerow(fieldnames)
        # write each result out
        for genome, genome_result in genome_results.items():

            values = [value for result in genome_result for value in result.values()]
            hit_subjects = [result['subject'] for result in genome_result]
            # add co-location result
            if len(set(hit_subjects)) == 1 and set(hit_subjects) != set(['NA']):
                row_items = [genome, 'Y'] # co-located
                row_items.extend(values)
                csv_writer.writerow(row_items)
            elif not write_only_co_located:
                row_items = [genome, 'N'] # not co-located
                row_items.extend(values)
                csv_writer.writerow(row_items)
    